#ifndef BOOST_MARSHAL_HPP
#define BOOST_MARSHAL_HPP

#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>
#include <boost/function.hpp>
#include <boost/type_traits/function_traits.hpp>
#include <boost/preprocessor/iterate.hpp>
#include <boost/preprocessor/repetition/enum_shifted_params.hpp>
#include <boost/ptr_container/ptr_map.hpp>
#include <boost/call_traits.hpp>
#include <sstream>
#include <boost/detail/arity.hpp>

namespace boost{
namespace marshall{

class function
{
public:
	void call(const std::string &params)
	{
		std::stringstream stream(params);
		boost::archive::binary_iarchive archive(stream);

		call(archive);
	}
	virtual void call(boost::archive::binary_iarchive &params) {};
};

namespace detail {

template<typename Signature, int Arity>
class marshalled_impl;

#define BOOST_PP_ITERATION_PARAMS_1 (3,(0,2,<boost/marshall/detail/marshalled_template.hpp>))
#include <boost/detail/arity_iterate.hpp>
#undef BOOST_PP_ITERATION_PARAMS_1

} // namespace detail

template<typename Signature>
class marshalled : public detail::marshalled_impl<Signature, boost::function_traits<Signature>::arity>
{
public:
	marshalled(const boost::function<Signature> &f)
		: detail::marshalled_impl<Signature, boost::function_traits<Signature>::arity>(f)
	{}
};

template<class Id>
class registry
{
public:
	template<typename Signature>
	// I'd like to use call traits param type for id but ptr_map complains on Id=int
	void set(Id id, const boost::function<Signature> &f)
	{
		reg.insert(id, new marshalled<Signature>(f));
	}
	void operator()(typename boost::call_traits<Id>::param_type id, const std::string &params)
	{
		reg[id].call(params);
	}
	void operator()(const std::string &id_and_params)
	{
		std::stringstream stream(id_and_params);
		boost::archive::binary_iarchive archive(stream);

		typename boost::detail::arity::storable<Id>::type id;
		archive & id;
		reg[id].call(archive);
	}
private:
	boost::ptr_map<Id, function> reg;
};

} // namespace marshall
} // namespace boost

#endif
